@@ -1,7 +1,7 @@
 #  File src/library/tools/R/testing.R
-#  Part of the R package, http://www.R-project.org
+#  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2014 The R Core Team
+#  Copyright (C) 1995-2022 The R Core Team
 #
 # NB: also copyright date in Usage.
 #
@@ -16,16 +16,17 @@
 #  GNU General Public License for more details.
 #
 #  A copy of the GNU General Public License is available at
-#  http://www.r-project.org/Licenses/
+#  https://www.R-project.org/Licenses/
 
 ## functions principally for testing R and packages
 
 massageExamples <-
     function(pkg, files, outFile = stdout(), use_gct = FALSE,
-             addTiming = FALSE, commentDonttest = TRUE)
+             addTiming = FALSE, ..., commentDonttest = TRUE)
 {
-    if(file_test("-d", files[1L])) {
-        old <- Sys.setlocale("LC_COLLATE", "C")
+    if(dir.exists(files[1L])) {
+        old <- Sys.getlocale("LC_COLLATE")
+        Sys.setlocale("LC_COLLATE", "C")
         files <- sort(Sys.glob(file.path(files, "*.R")))
         Sys.setlocale("LC_COLLATE", old)
     }
@@ -33,12 +34,18 @@ massageExamples <-
     if(is.character(outFile)) {
         out <- file(outFile, "wt")
         on.exit(close(out))
-    } else out <- outFile
+        cntFile <- paste0(outFile, "-cnt")
+    } else {
+        out <- outFile
+        cntFile <- NULL
+    }
+
+    count <- 0L # of files using \donttest
 
     lines <- c(paste0('pkgname <- "', pkg, '"'),
                'source(file.path(R.home("share"), "R", "examples-header.R"))',
                if (use_gct) {
-                   gct_n <- as.integer(Sys.getenv("_R_CHECK_GCT_N_", 0))
+                   gct_n <- as.integer(Sys.getenv("_R_CHECK_GCT_N_", "0"))
                    if(!is.na(gct_n) && gct_n > 0L)
                        sprintf("gctorture2(%s)", gct_n)
                    else "gctorture(TRUE)"
@@ -72,12 +79,14 @@ massageExamples <-
 
     cat("base::assign(\".oldSearch\", base::search(), pos = 'CheckExEnv')\n", file = out)
     ## cat("assign(\".oldNS\", loadedNamespaces(), pos = 'CheckExEnv')\n", file = out)
+    cat("base::assign(\".old_wd\", base::getwd(), pos = 'CheckExEnv')\n",
+        file = out)
     for(file in files) {
         nm <- sub("\\.R$", "", basename(file))
         ## make a syntactic name out of the filename
         nm <- gsub("[^- .a-zA-Z0-9_]", ".", nm, perl = TRUE, useBytes = TRUE)
         if (pkg == "grDevices" && nm == "postscript") next
-        ## Latin-1 examples are treated separat
+        ## Latin-1 examples are treated separately
         if (pkg == "graphics" && nm == "text") next
         if(!file.exists(file))
             stop("file ", file, " cannot be opened", domain = NA)
@@ -104,8 +113,10 @@ massageExamples <-
             dont_test <- FALSE
             for (line in lines) {
                 if(any(grepl("^[[:space:]]*## No test:", line,
-                             perl = TRUE, useBytes = TRUE)))
+                             perl = TRUE, useBytes = TRUE))) {
                     dont_test <- TRUE
+                    count <- count + 1L
+                }
                 if(!dont_test) cat(line, "\n", sep = "", file = out)
                 if(any(grepl("^[[:space:]]*## End\\(No test\\)", line,
                              perl = TRUE, useBytes = TRUE)))
@@ -127,11 +138,13 @@ massageExamples <-
 
     cat(readLines(file.path(R.home("share"), "R", "examples-footer.R")),
         sep = "\n", file = out)
+
+    if(count && !is.null(cntFile)) writeLines(as.character(count), cntFile)
 }
 
 ## compares 2 files
 Rdiff <- function(from, to, useDiff = FALSE, forEx = FALSE,
-                  nullPointers=TRUE, Log = FALSE)
+                  nullPointers = TRUE, Log = FALSE)
 {
     clean <- function(txt)
     {
@@ -141,30 +154,46 @@ Rdiff <- function(from, to, useDiff = FALSE, forEx = FALSE,
                               txt, perl = TRUE, useBytes = TRUE)) &&
            length(bot <- grep("quit R.$", txt, perl = TRUE, useBytes = TRUE)))
             txt <- txt[-(top[1L]:bot[1L])]
-        ## for massageExamples()
+        ## for massageExamples(), used for timings
         ll <- grep("</HEADER>", txt, fixed = TRUE, useBytes = TRUE)
         if(length(ll)) txt <- txt[-seq_len(max(ll))]
         ll <- grep("<FOOTER>", txt, fixed = TRUE, useBytes = TRUE)
         if(length(ll)) txt <- txt[seq_len(max(ll) - 1L)]
+        ## remove header change in R 3.5.0
+        if(forEx) {
+            ll <- grep('".old_wd"', txt, fixed = TRUE, useBytes = TRUE)
+            if(length(ll)) txt <- txt[-ll]
+        }
         ## remove BATCH footer
         nl <- length(txt)
-        if(nl > 3L && grepl("^> proc.time\\(\\)", txt[nl-2L])) txt <- txt[1:(nl-3L)]
-        if (nullPointers)
-        ## remove pointer addresses from listings
+        if(nl > 3L && startsWith(txt[nl-2L], "> proc.time()"))
+            txt <- txt[1:(nl-3L)]
+        ## remove text between IGNORE_RDIFF markers.
+        ## maybe this should only be done for forEx = TRUE?
+        txt <- txt[(cumsum(txt == "> ## IGNORE_RDIFF_BEGIN") <=
+                    cumsum(txt == "> ## IGNORE_RDIFF_END"))]
+        ## (Keeps the end markers, but that's ok.)
+        if (nullPointers) {
+            ## remove pointer addresses from listings
             txt <- gsub("<(environment|bytecode|pointer|promise): [x[:xdigit:]]+>", "<\\1: 0>", txt)
+            ## standardize hashtable, pro tem
+            txt <- sub("<hashtable.*>", "<hashtable output>", txt)
+        }
         ## regularize fancy quotes.  First UTF-8 ones:
-        txt <- gsub("(\xe2\x80\x98|\xe2\x80\x99)", "'", txt,
-                      perl = TRUE, useBytes = TRUE)
-        txt <- gsub("(\xe2\x80\x9c|\xe2\x80\x9d)", '"', txt,
-                      perl = TRUE, useBytes = TRUE)
+        txt <- .canonicalize_quotes(txt)
         if(.Platform$OS.type == "windows") {
             ## not entirely safe ...
-            txt <- gsub("(\x91|\x92)", "'", txt, perl = TRUE, useBytes = TRUE)
-            txt <- gsub("(\x93|\x94)", '"', txt, perl = TRUE, useBytes = TRUE)
-            txt <- txt[!grepl('options(pager = "console")', txt,
-                              fixed = TRUE, useBytes = TRUE)]
+            txt <- gsub(paste0("(",rawToChar(as.raw(0x91)),"|",rawToChar(as.raw(0x92)),")"),
+                        "'", txt, perl = TRUE, useBytes = TRUE)
+            txt <- gsub(paste0("(",rawToChar(as.raw(0x93)),"|",rawToChar(as.raw(0x94)),")"),
+                        '"', txt, perl = TRUE, useBytes = TRUE)
         }
-        pat <- '(^Time |^Loading required package|^Package [A-Za-z][A-Za-z0-9]+ loaded|^<(environment|promise|pointer|bytecode):|^/CreationDate |^/ModDate |^/Producer )'
+        ## massageExamples() adds options(pager = "console") only for
+        ## Windows, but we should ignore a corresponding diff on all
+        ## platforms.
+        txt <- txt[!grepl('options(pager = "console")', txt,
+                          fixed = TRUE, useBytes = TRUE)]
+        pat <- '(^Time |^Loading required package|^Package [A-Za-z][A-Za-z0-9]+ loaded|^<(environment|promise|pointer|bytecode):|^/CreationDate |^/ModDate |^/Producer |^End.Don\'t show)'
         txt[!grepl(pat, txt, perl = TRUE, useBytes = TRUE)]
     }
     clean2 <- function(txt)
@@ -178,16 +207,15 @@ Rdiff <- function(from, to, useDiff = FALSE, forEx = FALSE,
     if (forEx) {
         left <- clean2(left)
         ## remove lines from R CMD check --timings
-        left <- grep("[.](format_|)ptime", left, value = TRUE,
-                     invert = TRUE, useBytes = TRUE)
+        left <- filtergrep("[.](format_|)ptime", left, useBytes = TRUE)
         right <- clean2(right)
     }
     if (!useDiff && (length(left) == length(right))) {
         ## The idea is to emulate diff -b, as documented by POSIX:
-        ## http://pubs.opengroup.org/onlinepubs/9699919799/utilities/diff.html
-        bleft <- gsub("[[:space:]]*$", "", left)
+        ## https://pubs.opengroup.org/onlinepubs/9699919799/utilities/diff.html
+        bleft  <- gsub("[[:space:]]*$", "", left)
         bright <- gsub("[[:space:]]*$", "", right)
-        bleft <- gsub("[[:space:]]+", " ", bleft)
+        bleft  <- gsub("[[:space:]]+", " ", bleft)
         bright <- gsub("[[:space:]]+", " ", bright)
         if(all(bleft == bright))
             return(if(Log) list(status = 0L, out = character()) else 0L)
@@ -221,14 +249,37 @@ Rdiff <- function(from, to, useDiff = FALSE, forEx = FALSE,
             list(status = status, out = c(out, readLines(tf)))
         } else system(paste("diff -bw", shQuote(a), shQuote(b)))
     }
+} ## {Rdiff}
+
+.is.writeable <- function(dir)
+{
+    # see packages2.R for comment on unreliability of file.access
+    ok <- TRUE
+    fn <- file.path(dir, paste0("_test_dir_", Sys.getpid()))
+    res <- try(dir.create(fn, showWarnings = FALSE))
+    if(inherits(res, "try-error") || !res)
+        ok <- FALSE
+    else
+        unlink(fn, recursive = TRUE)
+    if (ok) {
+        fn <- file.path(dir, paste0("_test_file_", Sys.getpid()))
+        res <- try(file.create(fn, showWarnings = FALSE))
+        if(inherits(res, "try-error") || !res)
+            ok <- FALSE
+        else
+            unlink(fn)
+    }
+    ok
 }
 
 testInstalledPackages <-
     function(outDir = ".", errorsAreFatal = TRUE,
              scope = c("both", "base", "recommended"),
              types = c("examples", "tests", "vignettes"),
-             srcdir = NULL, Ropts = "")
+             srcdir = NULL, Ropts = "", ...)
 {
+    if (!.is.writeable(outDir))
+        stop("directory ", sQuote(outDir), " is not writeable ", domain = NA)
     ow <- options(warn = 1)
     on.exit(ow)
     scope <- match.arg(scope)
@@ -239,13 +290,13 @@ testInstalledPackages <-
         pkgs <- known_packages$base
     if (scope %in% c("both", "recommended"))
         pkgs <- c(pkgs, known_packages$recommended)
-    mc.cores <- as.integer(Sys.getenv("TEST_MC_CORES", 1L))
+    mc.cores <- as.integer(Sys.getenv("TEST_MC_CORES", "1"))
     if (.Platform$OS.type != "windows" &&
         !is.na(mc.cores) && mc.cores > 1L) {
         do_one <- function(pkg) {
             if(is.null(srcdir) && pkg %in% known_packages$base)
                 srcdir <- R.home("tests/Examples")
-            testInstalledPackage(pkg, .Library, outDir, types, srcdir, Ropts)
+            testInstalledPackage(pkg, .Library, outDir, types, srcdir, Ropts, ...)
         }
         res <- parallel::mclapply(pkgs, do_one, mc.cores = mc.cores,
                                   mc.preschedule = FALSE)
@@ -255,14 +306,16 @@ testInstalledPackages <-
                 warning(gettextf("testing '%s' failed", pkgs[i]),
                         domain = NA, call. = FALSE, immediate. = TRUE)
             if (errorsAreFatal)
-                stop(gettextf("%d of the package tests failed", sum(res)),
+                stop(sprintf(ngettext(sum(res), "%d of the package tests failed",
+                                      "%d of the package tests failed",
+                                       domain = "R-tools"), sum(res)),
                      domain = NA, call. = FALSE)
         }
     } else {
         for (pkg in pkgs) {
             if(is.null(srcdir) && pkg %in% known_packages$base)
                 srcdir <- R.home("tests/Examples")
-            res <- testInstalledPackage(pkg, .Library, outDir, types, srcdir, Ropts)
+            res <- testInstalledPackage(pkg, .Library, outDir, types, srcdir, Ropts, ...)
             if (res) {
                 status <- 1L
                 msg <- gettextf("testing '%s' failed", pkg)
@@ -277,22 +330,23 @@ testInstalledPackages <-
 testInstalledPackage <-
     function(pkg, lib.loc = NULL, outDir = ".",
              types = c("examples", "tests", "vignettes"),
-             srcdir = NULL, Ropts = "")
+             srcdir = NULL, Ropts = "", ...)
 {
-    types <- pmatch(types, c("examples", "tests", "vignettes"))
+    types <- match.arg(types, c("examples", "tests", "vignettes"), several.ok=TRUE)
     pkgdir <- find.package(pkg, lib.loc)
     owd <- setwd(outDir)
     on.exit(setwd(owd))
     strict <- as.logical(Sys.getenv("R_STRICT_PACKAGE_CHECK", "FALSE"))
+    useDiff <- nzchar(Sys.which("diff"))
 
-    if (1 %in% types) {
+    if ("examples" %in% types) {
         message(gettextf("Testing examples for package %s", sQuote(pkg)),
                 domain = NA)
-        Rfile <- .createExdotR(pkg, pkgdir, silent = TRUE)
+        Rfile <- .createExdotR(pkg, pkgdir, silent = TRUE, ...)
         if (length(Rfile)) {
             outfile <- paste0(pkg, "-Ex.Rout")
-            failfile <- paste(outfile, "fail", sep = "." )
-            savefile <- paste(outfile, "prev", sep = "." )
+            failfile <- paste0(outfile, ".fail")
+            savefile <- paste0(outfile, ".prev")
             if (file.exists(outfile)) file.rename(outfile, savefile)
             unlink(failfile)
             ## Create as .fail in case this R session gets killed
@@ -304,7 +358,7 @@ testInstalledPackage <-
             res <- system(cmd)
             if (res) return(invisible(1L)) else file.rename(failfile, outfile)
 
-            savefile <- paste(outfile, "save", sep = "." )
+            savefile <- paste0(outfile, ".save")
             if (!is.null(srcdir)) savefile <- file.path(srcdir, savefile)
             else {
                 tfile <- file.path(pkgdir, "tests", "Examples" , savefile)
@@ -316,19 +370,39 @@ testInstalledPackage <-
                    message(gettextf("  comparing %s to %s ...",
                                     sQuote(outfile), sQuote(basename(savefile))),
                            appendLF = FALSE, domain = NA)
-                    res <- Rdiff(outfile, savefile)
-                    if (!res) message(" OK")
-                    else if(strict)
-                        stop("  ", "results differ from reference results")
+                   cmd <-
+                       sprintf("invisible(tools::Rdiff('%s','%s',%s,TRUE))",
+                               outfile, savefile, as.character(useDiff))
+                   out <- R_runR(cmd, "--vanilla --no-echo")
+                   if(length(out)) {
+                       if(strict)
+                           message(" ERROR")
+                       else
+                           message(" NOTE")
+                       writeLines(paste0("  ", out))
+                       if(strict)
+                           stop("  ",
+                                "results differ from reference results")
+                   } else {
+                       message(" OK")
+                   }
                 }
             } else {
-                prevfile <- paste(outfile, "prev", sep = "." )
+                prevfile <- paste0(outfile, ".prev")
                 if (file.exists(prevfile)) {
                     message(gettextf("  comparing %s to %s ...",
                             sQuote(outfile), sQuote(basename(prevfile))),
                             appendLF = FALSE, domain = NA)
-                    res <- Rdiff(outfile, prevfile)
-                    if (!res) message(" OK")
+                    cmd <-
+                        sprintf("invisible(tools::Rdiff('%s','%s',%s,TRUE))",
+                                outfile, prevfile, as.character(useDiff))
+                    out <- R_runR(cmd, "--vanilla --no-echo")
+                    if(length(out)) {
+                        message(" NOTE")
+                        writeLines(paste0("  ", out))
+                    } else {
+                        message(" OK")
+                    }
                 }
             }
         } else
@@ -337,8 +411,8 @@ testInstalledPackage <-
     }
 
     ## FIXME merge with code in .runPackageTests
-    if (2 %in% types && file_test("-d", d <- file.path(pkgdir, "tests"))) {
-        this <- paste(pkg, "tests", sep = "-")
+    if ("tests" %in% types && dir.exists(d <- file.path(pkgdir, "tests"))) {
+        this <- paste0(pkg, "-tests")
         unlink(this, recursive = TRUE)
         dir.create(this)
         ## system(paste("cp -pR", file.path(d, "*"), this))
@@ -346,10 +420,10 @@ testInstalledPackage <-
         setwd(this)
         message(gettextf("Running specific tests for package %s",
                          sQuote(pkg)), domain = NA)
-        Rfiles <- dir(".", pattern="\\.R$")
+        Rfiles <- dir(".", pattern="\\.[rR]$")
         for(f in Rfiles) {
             message(gettextf("  Running %s", sQuote(f)), domain = NA)
-            outfile <- paste0(f, "out")
+            outfile <- sub("rout$", "Rout", paste0(f, "out"))
             cmd <- paste(shQuote(file.path(R.home("bin"), "R")),
                          "CMD BATCH --vanilla --no-timing", Ropts,
                          shQuote(f), shQuote(outfile))
@@ -357,10 +431,10 @@ testInstalledPackage <-
             else paste("LANGUAGE=C", cmd)
            res <- system(cmd)
             if (res) {
-                file.rename(outfile, paste(outfile, "fail", sep = "."))
+                file.rename(outfile, paste0(outfile, ".fail"))
                 return(invisible(1L))
             }
-            savefile <- paste(outfile, "save", sep = "." )
+            savefile <- paste0(outfile, ".save")
             if (file.exists(savefile)) {
                 message(gettextf("  comparing %s to %s ...",
                                  sQuote(outfile), sQuote(savefile)),
@@ -372,10 +446,11 @@ testInstalledPackage <-
         setwd(owd)
     }
 
-    if (3 %in% types && file_test("-d", d <- file.path(pkgdir, "doc"))) {
+    if ("vignettes" %in% types && dir.exists(file.path(pkgdir, "doc"))) {
         message(gettextf("Running vignettes for package %s", sQuote(pkg)),
                 domain = NA)
-        checkVignettes(pkg, lib.loc, latex = FALSE, weave =TRUE)
+        writeLines(format(checkVignettes(pkg, lib.loc = lib.loc,
+                                         latex = FALSE, weave = TRUE)))
     }
 
     invisible(0L)
@@ -392,11 +467,20 @@ testInstalledPackage <-
     q("no", status = status)
 }
 
-.runPackageTests <- function(use_gct = FALSE, use_valgrind = FALSE, Log = NULL)
+.runPackageTests <-
+    function(use_gct = FALSE, use_valgrind = FALSE, Log = NULL,
+             stop_on_error = TRUE, ...)
 {
+    tlim <- Sys.getenv("_R_CHECK_ONE_TEST_ELAPSED_TIMEOUT_",
+            Sys.getenv("_R_CHECK_TESTS_ELAPSED_TIMEOUT_",
+            Sys.getenv("_R_CHECK_ELAPSED_TIMEOUT_")))
+    tlim <- get_timeout(tlim)
     if (!is.null(Log)) Log <- file(Log, "wt")
     WINDOWS <- .Platform$OS.type == "windows"
     td0 <- as.numeric(Sys.getenv("_R_CHECK_TIMINGS_"))
+    theta <-
+        as.numeric(Sys.getenv("_R_CHECK_TEST_TIMING_CPU_TO_ELAPSED_THRESHOLD_",
+                              NA_character_))
     if (is.na(td0)) td0 <- Inf
     print_time <- function(t1, t2, Log)
     {
@@ -421,7 +505,7 @@ testInstalledPackage <-
                 appendLF = FALSE, domain = NA)
         if(!is.null(Log))
             cat("  Running ", sQuote(f), sep = "", file = Log)
-        outfile <- paste0(f, "out")
+        outfile <- sub("rout$", "Rout", paste0(f, "out"))
         cmd <- paste(shQuote(file.path(R.home("bin"), "R")),
                      "CMD BATCH --vanilla",
                      if(use_valgrind) "--debugger=valgrind",
@@ -432,14 +516,26 @@ testInstalledPackage <-
         } else
             cmd <- paste("LANGUAGE=C", "R_TESTS=startup.Rs", cmd)
         t1 <- proc.time()
-        res <- system(cmd)
+        res <- system(cmd, timeout = tlim)
         t2 <- proc.time()
         print_time(t1, t2, Log)
+        if (!WINDOWS && !is.na(theta)) {
+            td <- t2 - t1
+            cpu <- sum(td[-3L])
+            if(cpu >= pmax(theta * td[3L], 1)) {
+                ratio <- round(cpu/td[3L], 1L)
+                msg <- sprintf("Running R code in %s had CPU time %g times elapsed time\n",
+                               sQuote(f), ratio)
+                cat(msg)
+                if (!is.null(Log)) cat(msg, file = Log)
+            }
+        }
         if (res) {
-            file.rename(outfile, paste(outfile, "fail", sep = "."))
+            if(identical(res, 124L)) report_timeout(tlim)
+            file.rename(outfile, paste0(outfile, ".fail"))
             return(1L)
         }
-        savefile <- paste(outfile, "save", sep = "." )
+        savefile <- paste0(outfile, ".save")
         if (file.exists(savefile)) {
             message(gettextf("  Comparing %s to %s ...",
                              sQuote(outfile), sQuote(savefile)),
@@ -466,33 +562,31 @@ testInstalledPackage <-
     nfail <- 0L ## allow for later running all tests even if some fail.
     Rinfiles <- dir(".", pattern="\\.Rin$")
     for(f in Rinfiles) {
-        Rfile <- sub("\\.Rin$", ".R", f)
-        message("  Creating ", sQuote(Rfile), domain = NA)
+        message("  Processing ", sQuote(f), domain = NA)
         if (!is.null(Log))
-            cat("  Creating ", sQuote(Rfile), "\n", sep = "", file = Log)
+            cat("  Processing ", sQuote(f), "\n", sep = "", file = Log)
         cmd <- paste(shQuote(file.path(R.home("bin"), "R")),
-                     "CMD BATCH --no-timing --vanilla --slave", f)
+                     "CMD BATCH --no-timing --vanilla --no-echo", shQuote(f))
         if (system(cmd)) {
-            warning("creation of ", sQuote(Rfile), " failed", domain = NA)
-            if (!is.null(Log))
-                cat("Warning: creation of ", sQuote(Rfile), " failed\n",
-                    sep = "", file = Log)
-        } else if (file.exists(Rfile)) nfail <- nfail + runone(Rfile)
-        if (nfail > 0) return(nfail)
+            nfail <- nfail + 1L
+            file.rename(paste0(f, ".Rout"), paste0(f, ".Rout.fail"))
+            if (stop_on_error) return(1L)
+        }
     }
 
-    Rfiles <- dir(".", pattern="\\.R$")
+    Rfiles <- dir(".", pattern="\\.[rR]$")
     for(f in Rfiles) {
         nfail <- nfail + runone(f)
-        if (nfail > 0) return(nfail)
+        if (nfail > 0 && stop_on_error) return(nfail)
     }
     if (!is.null(Log)) close(Log)
     return(nfail)
 }
 
+## Defaults for commenting are the same as per-3.2.0 version.
 .createExdotR <-
     function(pkg, pkgdir, silent = FALSE, use_gct = FALSE, addTiming = FALSE,
-             commentDonttest = TRUE)
+             ..., commentDontrun = TRUE, commentDonttest = TRUE)
 {
     Rfile <- paste0(pkg, "-Ex.R")
 
@@ -513,8 +607,9 @@ testInstalledPackage <-
     for(f in files) {
         nm <- sub("\\.[Rr]d$", "", basename(f))
         Rd2ex(db[[f]],
-              file.path(filedir, paste(nm, "R", sep = ".")),
-              defines = NULL)
+              file.path(filedir, paste0(nm, ".R")),
+              defines = NULL, commentDontrun = commentDontrun,
+              commentDonttest = commentDonttest)
         cnt <- cnt + 1L
         if(!silent && cnt %% 10L == 0L)
             message(".", appendLF = FALSE, domain = NA)
@@ -524,41 +619,50 @@ testInstalledPackage <-
     if(!nof) return(invisible(NULL))
 
     massageExamples(pkg, filedir, Rfile, use_gct, addTiming,
-                    commentDonttest = commentDonttest)
+                    commentDonttest = commentDonttest, ...)
     invisible(Rfile)
 }
 
-testInstalledBasic <- function(scope = c("basic", "devel", "both"))
+testInstalledBasic <- function(scope = c("basic", "devel", "both", "internet"))
 {
     scope <- match.arg(scope)
 
     ## We need to force C collation: might not work
+    oLCcoll <- Sys.getlocale("LC_COLLATE") ; on.exit(Sys.setlocale("LC_COLLATE", oLCcoll))
     Sys.setlocale("LC_COLLATE", "C")
+    ## "strict specific":
     tests1 <- c("eval-etc", "simple-true", "arith-true", "lm-tests",
                 "ok-errors", "method-dispatch", "array-subset",
-                "any-all", "d-p-q-r-tests")
-    tests2 <- c("complex", "print-tests", "lapack", "datasets", "iec60559")
+                "p-r-random-tests", "d-p-q-r-tst-2",
+                "any-all", "structure", "d-p-q-r-tests")
+    ## "sloppy specific":
+    tests2 <- c("complex", "print-tests", "lapack", "datasets", "datetime",
+                "iec60559")
+    ## regression tests (strict specific, too)
     tests3 <- c("reg-tests-1a", "reg-tests-1b", "reg-tests-1c", "reg-tests-2",
+                "reg-tests-1d",
                 "reg-examples1", "reg-examples2", "reg-packages",
-                "reg-IO", "reg-IO2", "reg-S4", "reg-plot", "reg-BLAS")
+                "p-qbeta-strict-tst",
+                "reg-IO", "reg-IO2", "reg-plot", "reg-S4", "reg-BLAS")
 
+    useDiff <- nzchar(Sys.which("diff"))  # only check once
     runone <- function(f, diffOK = FALSE, inC = TRUE)
     {
-        f <- paste(f, "R", sep = ".")
+        f <- paste0(f, ".R")
         if (!file.exists(f)) {
             if (!file.exists(fin <- paste0(f, "in")))
                 stop("file ", sQuote(f), " not found", domain = NA)
             message("creating ", sQuote(f), domain = NA)
-            ## FIXME: this creates an extra trailing space compared to
-            ## the .Rin.R rule
             cmd <- paste(shQuote(file.path(R.home("bin"), "R")),
-                         "--vanilla --slave -f", fin)
+                         "--vanilla --no-echo -f", fin)
             if (system(cmd))
                 stop("creation of ", sQuote(f), " failed", domain = NA)
+            ## This needs an extra trailing space to match the .Rin.R rule
+            cat("\n", file = f, append = TRUE)
             on.exit(unlink(f))
         }
         message("  running code in ", sQuote(f), domain = NA)
-        outfile <- paste0(f, "out")
+        outfile <- sub("rout$", "Rout", paste0(f, "out"))
         cmd <- paste(shQuote(file.path(R.home("bin"), "R")),
                      "CMD BATCH --vanilla --no-timing",
                      shQuote(f), shQuote(outfile))
@@ -570,27 +674,38 @@ testInstalledBasic <- function(scope = c("basic", "devel", "both"))
             Sys.setenv(R_DEFAULT_PACKAGES="")
             Sys.setenv(LC_COLLATE="C")
             Sys.setenv(SRCDIR=".")
-            ## ignore inC and hope
+            ## FIXME: the above are currently not restored after testing
+            if (inC) { # breaks reg-plot-latin1, so restore between tests
+                oenv <- Sys.getenv("LC_CTYPE", unset = NA)
+                on.exit(if (is.na(oenv)) Sys.unsetenv("LC_CTYPE")
+                        else Sys.setenv(LC_CTYPE=oenv), add = TRUE)
+                Sys.setenv(LC_CTYPE="C")
+            }
+            ## ignore all 'extra' (incl. 'inC')  and hope
         } else cmd <- paste(extra, cmd)
         res <- system(cmd)
         if (res) {
-            file.rename(outfile, paste(outfile, "fail", sep = "."))
+            file.rename(outfile, paste0(outfile, ".fail"))
             message("FAILED")
             return(1L)
         }
-        savefile <- paste(outfile, "save", sep = "." )
+        savefile <- paste0(outfile, ".save")
         if (file.exists(savefile)) {
             message(gettextf("  comparing %s to %s ...",
                              sQuote(outfile), sQuote(savefile)),
                     appendLF = FALSE, domain = NA)
-            res <- Rdiff(outfile, savefile, TRUE)
+            res <- Rdiff(outfile, savefile, useDiff)
             if (!res) message(" OK")
             else if (!diffOK) return(1L)
         }
         0L
-    }
+    } # end{runone}
+
     owd <- setwd(file.path(R.home(), "tests"))
-    on.exit(setwd(owd))
+    on.exit(setwd(owd), add=TRUE)
+    if (!.is.writeable("."))
+        stop("directory ", sQuote(file.path(R.home(), "tests")),
+              " is not writeable ", domain = NA)
 
     if (scope %in% c("basic", "both")) {
         message("running strict specific tests", domain = NA)
@@ -607,30 +722,47 @@ testInstalledBasic <- function(scope = c("basic", "devel", "both"))
                 if(res != 0L) message("DIFFERED") else message("OK")
             }
         }
+        runone("reg-translation", inC=FALSE)
         runone("reg-tests-3", TRUE)
+        runone("reg-examples3", TRUE)
         message("running tests of plotting Latin-1", domain = NA)
         message("  expect failure or some differences if not in a Latin or UTF-8 locale", domain = NA)
 
-        runone("reg-plot-latin1", TRUE, FALSE)
-        message("  comparing 'reg-plot-latin1.pdf' to 'reg-plot-latin1.pdf.save' ...",
-                appendLF = FALSE, domain = NA)
-        res <- Rdiff("reg-plot-latin1.pdf", "reg-plot-latin1.pdf.save")
-        if(res != 0L) message("DIFFERED") else message("OK")
+        if (runone("reg-plot-latin1", TRUE, inC=FALSE) == 0L) {
+            message("  comparing 'reg-plot-latin1.pdf' to 'reg-plot-latin1.pdf.save' ...",
+                    appendLF = FALSE, domain = NA)
+            res <- Rdiff("reg-plot-latin1.pdf", "reg-plot-latin1.pdf.save")
+            if(res != 0L) message("DIFFERED") else message("OK")
+        }
     }
 
     if (scope %in% c("devel", "both")) {
+        message("running tests of date-time printing\n expect platform-specific differences", domain = NA)
+        runone("datetime2")
         message("running tests of consistency of as/is.*", domain = NA)
         runone("isas-tests")
         message("running tests of random deviate generation -- fails occasionally")
         runone("p-r-random-tests", TRUE)
+        message("running tests demos from base and stats", domain = NA)
+        if (runone("demos")) return(invisible(1L))
+        if (runone("demos2")) return(invisible(1L))
         message("running tests of primitives", domain = NA)
         if (runone("primitives")) return(invisible(1L))
         message("running regexp regression tests", domain = NA)
         if (runone("utf8-regex", inC = FALSE)) return(invisible(1L))
+        if (runone("PCRE")) return(invisible(1L))
+        message("running tests of CRAN tools", domain = NA)
+        if (runone("CRANtools")) return(invisible(1L))
         message("running tests to possibly trigger segfaults", domain = NA)
         if (runone("no-segfault")) return(invisible(1L))
     }
-
+    if (scope %in% "internet") {
+        message("running tests of Internet functions", domain = NA)
+        runone("internet")
+        message("running more Internet and socket tests", domain = NA)
+        runone("internet2")
+        runone("libcurl")
+    }
     invisible(0L)
 }
 
@@ -648,7 +780,7 @@ detachPackages <- function(pkgs, verbose = TRUE)
 
     ## The items need not all be packages
     ## and non-packages can be on the list multiple times.
-    isPkg <- grepl("^package:", pkgs)
+    isPkg <- startsWith(pkgs,"package:")
     for(item in pkgs[!isPkg]) {
         pos <- match(item, search())
         if(!is.na(pos)) .detach(pos)
@@ -664,24 +796,24 @@ detachPackages <- function(pkgs, verbose = TRUE)
     ## unloading 'grid' kills all devices
     ## tcltk is unhappy to have its DLL unloaded repeatedly
     exclusions <- c("grid", "tcltk")
-    exclusions <- paste("package", exclusions, sep = ":")
+    exclusions <- paste0("package:", exclusions)
     while(length(deps)) {
         unl <- unlist(deps)
         for(i in seq_along(deps)) {
             this <- names(deps)[i]
-            if(sub("^package:", "", this) %in% unl) next else break
+	    if(.rmpkg(this) %in% unl) next else break
         }
         ## hopefully force = TRUE is never needed, but it does ensure
         ## that progress gets made
         try(detach(this, character.only = TRUE,
-                   unload = unload && !(this %in% exclusions),
+                   unload = unload && (this %notin% exclusions),
                    force = TRUE))
         deps <- deps[-i]
     }
 }
 
-## Usage: Rscript --vanilla --default-packages=NULL args
-.Rdiff <- function()
+## Wrapper for  R CMD Rdiff   based on Rdiff() above :
+.Rdiff <- function(no.q = FALSE)
 {
     options(showErrorCalls=FALSE)
 
@@ -695,12 +827,16 @@ detachPackages <- function(pkgs, verbose = TRUE)
             "  -h, --help     print this help message and exit",
             "  -v, --version  print version info and exit",
             "",
-            "Report bugs at bugs.r-project.org .",
+            "Report bugs at <https://bugs.R-project.org>.",
             sep = "\n")
     }
 
-    do_exit <- function(status = 0L)
-        q("no", status = status, runLast = FALSE)
+    do_exit <-
+	if(no.q)
+	    function(status = 0L) (if(status) stop else message)(
+		".Rdiff() exit status ", status)
+	else
+	    function(status = 0L) q("no", status = status, runLast = FALSE)
 
     args <- commandArgs(TRUE)
     if (!length(args)) {
@@ -710,32 +846,33 @@ detachPackages <- function(pkgs, verbose = TRUE)
     args <- paste(args, collapse=" ")
     args <- strsplit(args,'nextArg', fixed = TRUE)[[1L]][-1L]
     if (length(args) == 1L) {
-        if(args[1L] %in% c("-h", "--help")) { Usage(); do_exit() }
+        if(args[1L] %in% c("-h", "--help")) { Usage(); do_exit(0) }
         if(args[1L] %in% c("-v", "--version")) {
             cat("R output diff: ",
                 R.version[["major"]], ".",  R.version[["minor"]],
                 " (r", R.version[["svn rev"]], ")\n", sep = "")
             cat("",
-                "Copyright (C) 2000-2013 The R Core Team.",
+                .R_copyright_msg(2000),
                 "This is free software; see the GNU General Public License version 2",
                 "or later for copying conditions.  There is NO warranty.",
                 sep = "\n")
-            do_exit()
+            do_exit(0)
         }
         Usage()
         do_exit(1L)
     }
 
-    if (length(args) < 2L) {
+    if (length(args) == 0L) {
         Usage()
         do_exit(1L)
     }
     exitstatus <- as.integer(args[3L])
-    if(is.na(exitstatus)) exitstatus <- 0L
-
+    if(is.na(exitstatus)) # default, also if length(args) == 2
+        exitstatus <- 0L
     left <- args[1L]
     if(left == "-") left <- "stdin"
     status <- Rdiff(left, args[2L], useDiff = TRUE)
     if(status) status <- exitstatus
     do_exit(status)
-}
+} ## .Rdiff()
+
