@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1998-2014   The R Core Team.
+ *  Copyright (C) 1998-2021   The R Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  */
 
 /* <UTF8>
@@ -43,6 +43,10 @@
 
 #include <rlocale.h> /* for btowc */
 
+#ifdef Win32
+#include <trioremap.h> /* for %lld */
+#endif
+
 /* The size of vector initially allocated by scan */
 #define SCAN_BLOCKSIZE		1000
 /* The size of the console buffer */
@@ -330,7 +334,7 @@ fillBuffer(SEXPTYPE type, int strip, int *bch, LocalData *d,
 */
     char *bufp;
     int c, quote, filled, nbuf = MAXELTSIZE, m, mm = 0;
-    Rboolean dbcslocale = (MB_CUR_MAX == 2);
+    Rboolean dbcslocale = (R_MB_CUR_MAX == 2) && !d->isUTF8 && !d->isLatin1;
 
     m = 0;
     filled = 1;
@@ -361,7 +365,7 @@ fillBuffer(SEXPTYPE type, int strip, int *bch, LocalData *d,
 		    buffer->data[m++] = (char) scanchar2(d);
 	    }
 	    if (c == R_EOF)
-	    	warning(_("EOF within quoted string"));
+		warning(_("EOF within quoted string"));
 	    c = scanchar(FALSE, d);
 	    mm = m;
 	}
@@ -411,7 +415,7 @@ fillBuffer(SEXPTYPE type, int strip, int *bch, LocalData *d,
 			    buffer->data[m++] = (char) scanchar2(d);
 		    }
 		    if (c == R_EOF)
-		    	warning(_("EOF within quoted string"));
+			warning(_("EOF within quoted string"));
 		    c = scanchar(TRUE, d); /* only peek at lead byte
 					      unless ASCII */
 		    if (c == quote) {
@@ -453,8 +457,8 @@ fillBuffer(SEXPTYPE type, int strip, int *bch, LocalData *d,
     }
     *bufp = '\0';
     /* Remove UTF-8 BOM */
-    if(d->atStart && utf8locale && 
-       !memcmp(buffer->data, "\xef\xbb\xbf", 3))
+    if(d->atStart && utf8locale &&
+       !strncmp(buffer->data, "\xef\xbb\xbf", 3))
 	memmove(buffer->data, buffer->data+3, strlen(buffer->data) + 1);
     d->atStart = FALSE;
     *bch = filled;
@@ -474,7 +478,7 @@ static R_INLINE int isNAstring(const char *buf, int mode, LocalData *d)
     return 0;
 }
 
-static R_INLINE void expected(char *what, char *got, LocalData *d)
+static R_INLINE void NORET expected(char *what, char *got, LocalData *d)
 {
     int c;
     if (d->ttyflag) { /* This is safe in a MBCS */
@@ -484,7 +488,7 @@ static R_INLINE void expected(char *what, char *got, LocalData *d)
     error(_("scan() expected '%s', got '%s'"), what, got);
 }
 
-static void extractItem(char *buffer, SEXP ans, int i, LocalData *d)
+static void extractItem(char *buffer, SEXP ans, R_xlen_t i, LocalData *d)
 {
     char *endp;
     switch(TYPEOF(ans)) {
@@ -546,11 +550,12 @@ static void extractItem(char *buffer, SEXP ans, int i, LocalData *d)
     }
 }
 
-static SEXP scanVector(SEXPTYPE type, int maxitems, int maxlines,
+static SEXP scanVector(SEXPTYPE type, R_xlen_t maxitems, R_xlen_t maxlines,
 		       int flush, SEXP stripwhite, int blskip, LocalData *d)
 {
     SEXP ans, bns;
-    int blocksize, c, i, n, linesread, nprev,strip, bch;
+    int c, strip, bch;
+    R_xlen_t i, blocksize, linesread, n, nprev;
     char *buffer;
     R_StringBuffer strBuf = {NULL, 0, MAXELTSIZE};
 
@@ -562,7 +567,8 @@ static SEXP scanVector(SEXPTYPE type, int maxitems, int maxlines,
 
     nprev = 0; n = 0; linesread = 0; bch = 1;
 
-    if (d->ttyflag) sprintf(ConsolePrompt, "1: ");
+    if (d->ttyflag)
+	snprintf(ConsolePrompt, CONSOLE_PROMPT_SIZE, "1: ");
 
     strip = asLogical(stripwhite);
 
@@ -576,13 +582,15 @@ static SEXP scanVector(SEXPTYPE type, int maxitems, int maxlines,
 	    linesread++;
 	    if (linesread == maxlines)
 		break;
-	    if (d->ttyflag) sprintf(ConsolePrompt, "%d: ", n + 1);
+	    if (d->ttyflag)
+		snprintf(ConsolePrompt, CONSOLE_PROMPT_SIZE,
+		         "%lld: ", (long long) (n + 1));
 	    nprev = n;
 	}
 	if (n == blocksize) {
 	    /* enlarge the vector*/
 	    bns = ans;
-	    if(blocksize > INT_MAX/2) error(_("too many items"));
+	    if(blocksize > R_XLEN_T_MAX/2) error(_("too many items"));
 	    blocksize = 2 * blocksize;
 	    ans = allocVector(type, blocksize);
 	    UNPROTECT(1);
@@ -610,7 +618,8 @@ static SEXP scanVector(SEXPTYPE type, int maxitems, int maxlines,
 	    bch = c;
 	}
     }
-    if (!d->quiet) REprintf("Read %d item%s\n", n, (n == 1) ? "" : "s");
+    if (!d->quiet)
+	REprintf("Read %lld item%s\n", (long long) n, (n == 1) ? "" : "s");
     if (d->ttyflag) ConsolePrompt[0] = '\0';
 
     if (n == 0) {
@@ -656,17 +665,18 @@ static SEXP scanVector(SEXPTYPE type, int maxitems, int maxlines,
 }
 
 
-static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
-		      int fill, SEXP stripwhite, int blskip, int multiline,
-		      LocalData *d)
+static SEXP scanFrame(SEXP what, R_xlen_t maxitems, R_xlen_t maxlines,
+                      int flush, int fill, SEXP stripwhite, int blskip,
+                      int multiline, LocalData *d)
 {
     SEXP ans, new, old, w;
     char *buffer = NULL;
-    int blksize, c, i, ii, j, n, nc, linesread, colsread, strip, bch;
-    int badline, nstring = 0;
+    int c, strip, bch;
+    R_xlen_t blksize, i, ii, j, n, nc, linesread, colsread;
+    R_xlen_t badline, nstring = 0;
     R_StringBuffer buf = {NULL, 0, MAXELTSIZE};
 
-    nc = length(what);
+    nc = xlength(what);
     if (!nc) {
 	    error(_("empty 'what' specified"));
     }
@@ -694,9 +704,13 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
     bch = 1;
     c = 0;			/* -Wall */
 
-    if (d->ttyflag) sprintf(ConsolePrompt, "1: ");
+    if (d->ttyflag)
+	snprintf(ConsolePrompt, CONSOLE_PROMPT_SIZE, "1: ");
 
-    strip = asLogical(stripwhite);
+    // we checked its type in do_scan
+    int *lstrip = LOGICAL(stripwhite);
+    Rboolean vec_strip = (xlength(stripwhite) == xlength(what));
+    strip = lstrip[0];
 
     for (;;) {
 	if(linesread % 1000 == 999) R_CheckUserInterrupt();
@@ -719,17 +733,19 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
 		} else if (!badline && !multiline)
 		    badline = linesread;
 		if(badline && !multiline)
-		error(_("line %d did not have %d elements"), badline, nc);
+		error(_("line %lld did not have %lld elements"),
+		    (long long) badline, (long long) nc);
 	    }
 	    if (maxitems > 0 && n >= maxitems)
 		goto done;
 	    if (maxlines > 0 && linesread == maxlines)
 		goto done;
 	    if (d->ttyflag)
-		sprintf(ConsolePrompt, "%d: ", n + 1);
+		snprintf(ConsolePrompt, CONSOLE_PROMPT_SIZE,
+		         "%lld: ", (long long) (n + 1));
 	}
 	if (n == blksize && colsread == 0) {
-	    if(blksize > INT_MAX/2) error(_("too many items"));
+	    if(blksize > R_XLEN_T_MAX/2) error(_("too many items"));
 	    blksize = 2 * blksize;
 	    for (i = 0; i < nc; i++) {
 		old = VECTOR_ELT(ans, i);
@@ -741,6 +757,7 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
 	    }
 	}
 
+	if (vec_strip) strip = lstrip[colsread];
 	buffer = fillBuffer(TYPEOF(VECTOR_ELT(ans, ii)), strip, &bch, d, &buf);
 	if (colsread == 0 &&
 	    strlen(buffer) == 0 &&
@@ -752,8 +769,6 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
 	    extractItem(buffer, VECTOR_ELT(ans, ii), n, d);
 	    ii++;
 	    colsread++;
-	    if (length(stripwhite) == length(what))
-		strip = LOGICAL(stripwhite)[colsread];
 	    /* increment n and reset i after filling a row */
 	    if (colsread == nc) {
 		n++;
@@ -763,8 +778,6 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
 		    while ((c = scanchar(FALSE, d)) != '\n' && c != R_EOF);
 		    bch = c;
 		}
-		if (length(stripwhite) == length(what))
-		    strip = LOGICAL(stripwhite)[0];
 	    }
 	}
     }
@@ -779,7 +792,8 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
 	}
 	n++;
     }
-    if (!d->quiet) REprintf("Read %d record%s\n", n, (n == 1) ? "" : "s");
+    if (!d->quiet)
+	REprintf("Read %lld record%s\n", (long long) n, (n == 1) ? "" : "s");
     if (d->ttyflag) ConsolePrompt[0] = '\0';
 
     for (i = 0; i < nc; i++) {
@@ -822,24 +836,24 @@ static SEXP scanFrame(SEXP what, int maxitems, int maxlines, int flush,
 SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, file, sep, what, stripwhite, dec, quotes, comstr;
-    int i, c, nlines, nmax, nskip, flush, fill, blskip, multiline, 
-	escapes, skipNul;
+    int c, flush, fill, blskip, multiline, escapes, skipNul;
+    R_xlen_t nmax, nlines, nskip;
     const char *p, *encoding;
     RCNTXT cntxt;
     LocalData data = {NULL, 0, 0, '.', NULL, NO_COMCHAR, 0, NULL, FALSE,
-		      FALSE, 0, FALSE, FALSE, FALSE, FALSE};
+		      FALSE, 0, FALSE, FALSE, FALSE, FALSE, FALSE, {FALSE}};
     data.NAstrings = R_NilValue;
 
     checkArity(op, args);
 
     file = CAR(args);		   args = CDR(args);
     what = CAR(args);		   args = CDR(args);
-    nmax = asInteger(CAR(args));   args = CDR(args);
+    nmax = asXLength(CAR(args));   args = CDR(args);
     sep = CAR(args);		   args = CDR(args);
     dec = CAR(args);		   args = CDR(args);
     quotes = CAR(args);		   args = CDR(args);
-    nskip = asInteger(CAR(args));  args = CDR(args);
-    nlines = asInteger(CAR(args)); args = CDR(args);
+    nskip = asXLength(CAR(args));  args = CDR(args);
+    nlines = asXLength(CAR(args)); args = CDR(args);
     data.NAstrings = CAR(args);	   args = CDR(args);
     flush = asLogical(CAR(args));  args = CDR(args);
     fill  = asLogical(CAR(args));  args = CDR(args);
@@ -859,13 +873,13 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
     if (data.quiet == NA_LOGICAL)		data.quiet = 0;
     if (blskip == NA_LOGICAL)			blskip = 1;
     if (multiline == NA_LOGICAL)		multiline = 1;
-    if (nskip < 0 || nskip == NA_INTEGER)	nskip = 0;
-    if (nlines < 0 || nlines == NA_INTEGER)	nlines = 0;
-    if (nmax < 0 || nmax == NA_INTEGER)		nmax = 0;
+    if (nskip < 0)				nskip = 0;
+    if (nlines < 0)				nlines = 0;
+    if (nmax < 0)				nmax = 0;
 
     if (TYPEOF(stripwhite) != LGLSXP)
 	error(_("invalid '%s' argument"), "strip.white");
-    if (length(stripwhite) != 1 && length(stripwhite) != length(what))
+    if (xlength(stripwhite) != 1 && xlength(stripwhite) != xlength(what))
 	error(_("invalid 'strip.white' length"));
     if (TYPEOF(data.NAstrings) != STRSXP)
 	error(_("invalid '%s' argument"), "na.strings");
@@ -898,7 +912,7 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
 
     if (isString(quotes)) {
 	const char *sc = translateChar(STRING_ELT(quotes, 0));
-	if (strlen(sc)) data.quoteset = strdup(sc);
+	if (strlen(sc)) data.quoteset = Rstrdup(sc);
 	else data.quoteset = "";
     } else if (isNull(quotes))
 	data.quoteset = "";
@@ -917,9 +931,9 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
 	error(_("invalid '%s' argument"), "skipNul");
     data.skipNul = skipNul != 0;
 
-    i = asInteger(file);
-    data.con = getConnection(i);
-    if(i == 0) {
+    int ii = asInteger(file);
+    data.con = getConnection(ii);
+    if(ii == 0) {
 	data.atStart = FALSE;
 	data.ttyflag = 1;
     } else {
@@ -936,10 +950,10 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
 		error(_("cannot read from this connection"));
 	    }
 	} else {
-	    if(!data.con->canread) 
+	    if(!data.con->canread)
 		error(_("cannot read from this connection"));
 	}
-	for (i = 0; i < nskip; i++) /* MBCS-safe */
+	for (R_xlen_t i = 0; i < nskip; i++) /* MBCS-safe */
 	    while ((c = scanchar(FALSE, &data)) != '\n' && c != R_EOF);
     }
 
@@ -948,8 +962,8 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
 
     /* set up a context which will close the connection if there is
        an error or user interrupt */
-    begincontext(&cntxt, CTXT_CCODE, call, R_BaseEnv, R_BaseEnv,
-		 R_NilValue, R_NilValue);
+    begincontext(&cntxt, CTXT_CCODE, R_GlobalContext->call, R_BaseEnv,
+		 R_BaseEnv, R_NilValue, R_NilValue);
     cntxt.cend = &scan_cleanup;
     cntxt.cenddata = &data;
 
@@ -971,6 +985,7 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
     default:
 	error(_("invalid '%s' argument"), "what");
     }
+    PROTECT(ans);
     endcontext(&cntxt);
 
     /* we might have a character that was unscanchar-ed.
@@ -983,8 +998,10 @@ SEXP attribute_hidden do_scan(SEXP call, SEXP op, SEXP args, SEXP rho)
     if (!data.ttyflag && !data.wasopen)
 	data.con->close(data.con);
     if (data.quoteset[0]) free(data.quoteset);
-    if (!skipNul && data.embedWarn) 
+    if (!skipNul && data.embedWarn)
 	warning(_("embedded nul(s) found in input"));
+
+    UNPROTECT(1); /* ans */
     return ans;
 }
 
@@ -1005,8 +1022,8 @@ SEXP attribute_hidden do_readln(SEXP call, SEXP op, SEXP args, SEXP rho)
 	if(length(prompt) > 0) {
 	    strncpy(ConsolePrompt, translateChar(STRING_ELT(prompt, 0)),
 		    CONSOLE_PROMPT_SIZE - 1);
-            ConsolePrompt[CONSOLE_PROMPT_SIZE - 1] = '\0';
-        }
+	    ConsolePrompt[CONSOLE_PROMPT_SIZE - 1] = '\0';
+	}
     }
 
     if(R_Interactive) {
